// errorcheck

// Copyright 2021 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package main

func main() {
	ch := make(chan struct{ v [65536]byte }) // ERROR "channel element type too large"
	close(ch)
}
