// Copyright 2021 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package main

import (
	"./b"
	"fmt"
)

func main() {
	opt := b.Ap1[string, string]{}
	fmt.Println(opt)
}
