// Copyright 2021 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package a

type S struct{}

func (s *S) M() {
	s.m((*S).N)
}

func (s *S) N() {}

func (s *S) m(func(*S)) {}
