// Copyright 2013 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package a

type unexported struct {
        a int
        b bool
}

type Struct struct {
        unexported
}
