// rundir

// Copyright 2022 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Test that an embedded unexported type has a different name in
// different packages.

package ignored
