// errorcheck -std

// Copyright 2022 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package p

//go:uintptrkeepalive
func missingNosplit(uintptr) { // ERROR "go:uintptrkeepalive requires go:nosplit"
}
