// Copyright 2022 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package b

import "./a"

func f() {
	a.Append()
	a.Delete()
	a.Print()
	a.Println()
	a.Complex()
	a.Copy()
	a.UnsafeAdd()
	a.UnsafeSlice()
}
