// compiledir

// Copyright 2021 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// This test exercises exporting + importing method
// expressions for use when inlining.

package ignored
